//Project: WinchBot
//Homepage: www.HomoFaciens.de
//Author Norbert Heinz
//Version: 0.1
//Creation date: 14.02.2017
//This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//For a copy of the GNU General Public License see http://www.gnu.org/licenses/
//
//compile with gcc WinchBot.c -o WinchBot -lm
//For details see:
//http://www.HomoFaciens.de/technics-machines-winchbot_en.htm

#include <stdio.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>



#define BAUDRATE B115200
//#define ARDUINOPORT "/dev/ttyUSB0"
#define ARDUINOPORT "/dev/ttyACM0"
#define FALSE 0
#define TRUE 1

//Base length of triangle in mm
#define BASELENGTH          895.0
//Initial cord length in mm
#define INITIAL_LENGTH       600.0
//Distance mount point cords to center of ball bearing mount at triangle
#define INITIAL_BAR_LENGTH   373.0
//Mount point cords on square tube
#define RADIUS_CORD_MOUNT     23.0
//Height from pivot point servo 4 to gripper
#define LEVER_S5 50.0
//Horizontal distance from pivot point servo 4 to tip of gripper
#define GRIPPER_LENGTH 60.0
//Horizontal offset from pivot point Servo 3 to centerline gripper
#define GRIPPER_OFFSET  5.0

#define PI 3.1415927

#define SERVO_1_MIN   45
#define SERVO_1_MAX  135

#define SERVO_2_MIN   45
#define SERVO_2_MAX  135

//Center pos of Servo3 is 76 in Arduino code
//Min = 90 - 76
//Max = 180 + Min
#define SERVO_3_MIN   24
#define SERVO_3_MAX  204

//Center pos of Servo4 is 47 in Arduino code
//Min = 90 - 47
//Max = 180 + Min
#define SERVO_4_MIN   43
#define SERVO_4_MAX  227

#define SERVO_5_MIN    0
#define SERVO_5_MAX  180

double StepsPermm = 360.0 / 162.0;

double length1, length2, length3;


double M1X, M1Y, M1Z;
double M2X, M2Y, M2Z;
double M3X, M3Y, M3Z;

int Servo1 = 90;
int Servo2 = 90;
int Servo3 = 90;
int Servo4 = 90;
int Servo5 = 90;

long currentX = 0, currentY = 0, currentZ = 0;

int initBot = 0;

int KeyHit = 0;
int KeyCode[5];
int SingleKey=0;


int  MaxRows = 24;
int  MaxFileRows = 10;
int  FilesFound = 0;
int  MaxCols = 80;
int  MessageX = 1;
int  MessageY = 24;
double Scale = 1.0;
long MoveLength = 1;
long xMin = 1000000, xMax = -1000000;
long yMin = 1000000, yMax = -1000000;
char FileName[1000] = "";

long StepX = 0;
long StepY = 0;
long StepZ = 0;

int fd = 0;
struct termios TermOpt;
int waitForKeyHit = 1;


char ReplayPath[1000];




//+++++++++++++++++++++++ Start gotoxy ++++++++++++++++++++++++++
//Thanks to 'Stack Overflow', found on http://www.daniweb.com/software-development/c/code/216326
int gotoxy(int x, int y) {
  char essq[100]; // String variable to hold the escape sequence
  char xstr[100]; // Strings to hold the x and y coordinates
  char ystr[100]; // Escape sequences must be built with characters

  //Convert the screen coordinates to strings.
  sprintf(xstr, "%d", x);
  sprintf(ystr, "%d", y);

  //Build the escape sequence (vertical move).
  essq[0] = '\0';
  strcat(essq, "\033[");
  strcat(essq, ystr);

  //Described in man terminfo as vpa=\E[%p1%dd. Vertical position absolute.
  strcat(essq, "d");

  //Horizontal move. Horizontal position absolute
  strcat(essq, "\033[");
  strcat(essq, xstr);
  // Described in man terminfo as hpa=\E[%p1%dG
  strcat(essq, "G");

  //Execute the escape sequence. This will move the cursor to x, y
  printf("%s", essq);
  return 0;
}
//------------------------ End gotoxy ----------------------------------

//+++++++++++++++++++++++ Start clrscr ++++++++++++++++++++++++++
void clrscr(int StartRow, int EndRow) {
  int i, i2;

  if (EndRow < StartRow){
    i = EndRow;
    EndRow = StartRow;
    StartRow = i;
  }
  gotoxy(1, StartRow);
  for (i = 0; i <= EndRow - StartRow; i++){
    for(i2 = 0; i2 < MaxCols; i2++){
      printf(" ");
    }
    printf("\n");
  }
}
//----------------------- End clrscr ----------------------------

//+++++++++++++++++++++++ Start kbhit ++++++++++++++++++++++++++++++++++
//Thanks to Undertech Blog, http://www.undertec.de/blog/2009/05/kbhit_und_getch_fur_linux.html
int kbhit(void) {

   struct termios term, oterm;
   int fd = 0;
   int c = 0;

   tcgetattr(fd, &oterm);
   memcpy(&term, &oterm, sizeof(term));
   term.c_lflag = term.c_lflag & (!ICANON);
   term.c_cc[VMIN] = 0;
   term.c_cc[VTIME] = waitForKeyHit;
   tcsetattr(fd, TCSANOW, &term);
   c = getchar();
   tcsetattr(fd, TCSANOW, &oterm);
   if (c != -1)
   ungetc(c, stdin);

   return ((c != -1) ? 1 : 0);

}
//------------------------ End kbhit -----------------------------------

//+++++++++++++++++++++++ Start getch ++++++++++++++++++++++++++++++++++
//Thanks to Undertech Blog, http://www.undertec.de/blog/2009/05/kbhit_und_getch_fur_linux.html
int getch(){
   static int ch = -1, fd = 0;
   struct termios new, old;

   fd = fileno(stdin);
   tcgetattr(fd, &old);
   new = old;
   new.c_lflag &= ~(ICANON|ECHO);
   tcsetattr(fd, TCSANOW, &new);
   ch = getchar();
   tcsetattr(fd, TCSANOW, &old);

//   printf("ch=%d ", ch);

   return ch;
}
//------------------------ End getch -----------------------------------

//++++++++++++++++++++++ Start MessageText +++++++++++++++++++++++++++++
void MessageText(char *message, int x, int y, int alignment){
  int i;
  char TextLine[300];

  clrscr(y, y);
  gotoxy (x, y);

  TextLine[0] = '\0';
  if(alignment == 1){
    for(i=0; i < (MaxCols - strlen(message)) / 2 ; i++){
      strcat(TextLine, " ");
    }
  }
  strcat(TextLine, message);

  printf("%s\n", TextLine);
}
//-------------------------- End MessageText ---------------------------

//++++++++++++++++++++++ Start PrintRow ++++++++++++++++++++++++++++++++
void PrintRow(char character, int y){
  int i;
  gotoxy (1, y);
  for(i=0; i<MaxCols;i++){
    printf("%c", character);
  }
}
//-------------------------- End PrintRow ------------------------------

//+++++++++++++++++++++++++ ErrorText +++++++++++++++++++++++++++++
void ErrorText(char *message){
  clrscr(MessageY + 2, MessageY + 2);
  gotoxy (1, MessageY + 2);
  printf("Last error: %s", message);
}
//----------------------------- ErrorText ---------------------------


void PrintYN(){
   clrscr(1, MessageY);
   gotoxy(1, 1);
   MessageText(" Shutdown Winchbot?", 1, 1, 1);
   MessageText("  -------------", 1, 3, 1);
   MessageText("  |     |     |", 1, 4, 1);
   MessageText("  |  Y  |  N  |", 1, 5, 1);
   MessageText("  |     |     |", 1, 6, 1);
   MessageText("  -------------", 1, 7, 1);
}

//+++++++++++++++++++++++++ PrintMenue_01 ++++++++++++++++++++++++++++++
void PrintMenue_01(){
  char TextLine[300];

   clrscr(1, MessageY-2);
   MessageText("*** Main menu plotter ***", 1, 1, 1);
   sprintf(TextLine, "M            - toggle move length, current value = %ld step(s)", MoveLength);
   MessageText(TextLine, 10, 2, 0);
   MessageText("Cursor right - move bot in positive X direction", 10, 3, 0);
   MessageText("Cursor left  - move bot in negative X direction", 10, 4, 0);
   MessageText("Cursor up    - move bot in positive Y direction", 10, 5, 0);
   MessageText("Cursor down  - move bot in negative Y direction", 10, 6, 0);
   MessageText("Page up      - move bot in positive Z direction", 10, 7, 0);
   MessageText("Page down    - move bot in negative Z direction", 10, 8, 0);
   MessageText("L, R         - Rotate gripper Left/Right", 10, 9, 0);
   MessageText("U, D         - Gripper Up/Down", 10, 10, 0);
   MessageText("O, C         - Open/Close gripper", 10, 11, 0);
   if(initBot == 0){
     sprintf(TextLine, "I            - set initial length of cords");
   }
   else{
     sprintf(TextLine, "I            - initial length of cords finished");
   }
   MessageText(TextLine, 10, 9, 0);

   sprintf(TextLine, "F            - choose file: \"%s\"", FileName);
   MessageText(TextLine, 10, 12, 0);
   sprintf(TextLine, "S            - Add X/Y/Z to Coordinates.bot X=%ld, Y=%ld, Z=%ld, S3=%d, S4=%d, S5=%d", currentX, currentY, currentZ, Servo3, Servo4, Servo5);
   MessageText(TextLine, 10, 13, 0);
   MessageText("P            - process file", 10, 14, 0);

   MessageText("Esc          - leave program", 10, 15, 0);

   PrintRow('-', MessageY - 1);


}
//------------------------- PrintMenue_01 ------------------------------

//+++++++++++++++++++++++++ PrintMenue_02 ++++++++++++++++++++++++++++++
char *PrintMenue_02(int StartRow, int selected){
  char TextLine[300];
  char OpenDirName[1000];
  static char FileName[101];
  DIR *pDIR;
  struct dirent *pDirEnt;
  int i = 0;
  int Discard = 0;

  clrscr(1, MessageY-2);
  MessageText("*** Choose plotter file ***", 1, 1, 1);

  strcpy(OpenDirName, ReplayPath);


  pDIR = opendir(OpenDirName);
  if ( pDIR == NULL ) {
    sprintf(TextLine, "Could not open directory '%s'!", OpenDirName);
    MessageText(TextLine, 1, 4, 1);
    getch();
    return( "" );
  }

  FilesFound = 0;
  pDirEnt = readdir( pDIR );
  while ( pDirEnt != NULL && i < MaxFileRows) {
    //printf("%s\n", pDirEnt->d_name);
    if(strlen(pDirEnt->d_name) > 4){
      if(memcmp(pDirEnt->d_name + strlen(pDirEnt->d_name)-4, ".bot",4) == 0){
        FilesFound++;
        if(Discard >= StartRow){
          if(i + StartRow == selected){
            sprintf(TextLine, ">%s<", pDirEnt->d_name);
            strcpy(FileName, pDirEnt->d_name);
          }
          else{
            sprintf(TextLine, " %s ", pDirEnt->d_name);
          }
          MessageText(TextLine, 1, 3 + i, 0);
          i++;
        }
        Discard++;
      }
    }
    pDirEnt = readdir( pDIR );
  }

  gotoxy(MessageX, MessageY + 1);
  printf("Choose file using up/down keys and confirm with 'Enter' or press 'Esc' to cancel.");


  return (FileName);
}
//------------------------- PrintMenue_02 ------------------------------


//+++++++++++++++++++++++++ PrintMenue_03 ++++++++++++++++++++++++++++++
void PrintMenue_03(char *FullFileName, long CurrentLine, long StartTime){
  char TextLine[300];
  long CurrentTime, ProcessHours = 0, ProcessMinutes = 0, ProcessSeconds = 0;

   CurrentTime = time(0);

   CurrentTime -= StartTime;

   while (CurrentTime > 3600){
     ProcessHours++;
     CurrentTime -= 3600;
   }
   while (CurrentTime > 60){
     ProcessMinutes++;
     CurrentTime -= 60;
   }
   ProcessSeconds = CurrentTime;

   clrscr(1, MessageY - 2);
   MessageText("*** Plotting file ***", 1, 1, 1);

   sprintf(TextLine, "File name: %s", FullFileName);
   MessageText(TextLine, 10, 3, 0);
   sprintf(TextLine, "Point No. %ld: X=%ld, Y=%ld, Z=%ld, S3=%d, S4=%d, S5=%d     ", CurrentLine, currentX, currentY, currentZ, Servo3, Servo4, Servo5);
   MessageText(TextLine, 10, 5, 0);
   sprintf(TextLine, "Process time: %02ld:%02ld:%02ld", ProcessHours, ProcessMinutes, ProcessSeconds);
   MessageText(TextLine, 10, 6, 0);


}
//------------------------- PrintMenue_03 ------------------------------


//+++++++++++++++++++++++ Start readport ++++++++++++++++++++++++++
char  readport(void){
  int n;
  char buff;


  n = read(fd, &buff, 1);
  if(n > 0){
    return buff;
  }

  return 0;
}
//------------------------ End readport ----------------------------------

//+++++++++++++++++++++++ Start sendport ++++++++++++++++++++++++++
void sendport(unsigned char ValueToSend){
  int n;

  n = write(fd, &ValueToSend, 1);

  if (n < 0){
    ErrorText("write() of value failed!\r");
  }
  else{
    while(readport() != 'r');
  }

}
//------------------------ End sendport ----------------------------------

//+++++++++++++++++++++++ Start openport ++++++++++++++++++++++++++
void openport(void){

    fd = open(ARDUINOPORT, O_RDWR | O_NOCTTY | O_NDELAY);
    if (fd == -1)  {
      ErrorText("init_serialport: Unable to open port ");
    }

    if (tcgetattr(fd, &TermOpt) < 0) {
      ErrorText("init_serialport: Couldn't get term attributes");
    }
    speed_t brate = BAUDRATE; // let you override switch below if needed

    cfsetispeed(&TermOpt, brate);
    cfsetospeed(&TermOpt, brate);

    // 8N1
    TermOpt.c_cflag &= ~PARENB;
    TermOpt.c_cflag &= ~CSTOPB;
    TermOpt.c_cflag &= ~CSIZE;
    TermOpt.c_cflag |= CS8;
    // no flow control
    TermOpt.c_cflag &= ~CRTSCTS;

    TermOpt.c_cflag |= CREAD | CLOCAL;  // turn on READ & ignore ctrl lines
    TermOpt.c_iflag &= ~(IXON | IXOFF | IXANY); // turn off s/w flow ctrl

    TermOpt.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // make raw
    TermOpt.c_oflag &= ~OPOST; // make raw

    // see: http://unixwiz.net/techtips/termios-vmin-vtime.html
    TermOpt.c_cc[VMIN]  = 0;
    TermOpt.c_cc[VTIME] = 20;

    if( tcsetattr(fd, TCSANOW, &TermOpt) < 0) {
      ErrorText("init_serialport: Couldn't set term attributes");
    }

}
//------------------------ End openport ----------------------------------


//++++++++++++++++++++++++++++++ move servo 1 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void moveServo1(int direction){

  if((Servo1 < SERVO_1_MAX && direction > 0) || (Servo1 > SERVO_1_MIN && direction < 0)){
    Servo1 += direction;
    if(direction > 0){
      sendport('g');
    }
    else{
      sendport('h');
    }
  }
}

//++++++++++++++++++++++++++++++ move servo 2 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void moveServo2(int direction){

  if((Servo2 < SERVO_2_MAX && direction > 0) || (Servo2 > SERVO_2_MIN && direction < 0)){
    Servo2 += direction;
    if(direction > 0){
      sendport('i');
    }
    else{
      sendport('j');
    }
  }
}

//++++++++++++++++++++++++++++++ move servo 3 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void moveServo3(int direction){

  if((Servo3 < SERVO_3_MAX && direction > 0) || (Servo3 > SERVO_3_MIN && direction < 0)){
    Servo3 += direction;
    if(direction > 0){
      sendport('k');
    }
    else{
      sendport('l');
    }
  }
}

//++++++++++++++++++++++++++++++ move servo 4 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void moveServo4(int direction){

  if((Servo4 < SERVO_4_MAX && direction > 0) || (Servo4 > SERVO_4_MIN && direction < 0)){
    Servo4 += direction;
    if(direction > 0){
      sendport('m');
    }
    else{
      sendport('n');
    }
  }
}

//++++++++++++++++++++++++++++++ move servo 5 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void moveServo5(int direction){

  if((Servo5 < SERVO_5_MAX && direction > 0) || (Servo5 > SERVO_5_MIN && direction < 0)){
    Servo5 += direction;
    if(direction > 0){
      sendport('o');
    }
    else{
      sendport('p');
    }
  }
}

//++++++++++++++++++++++++++++++ move motor 1 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void motor1(int direction){
  length1 += direction;

  if(direction < 0){
    sendport('a');
  }
  else{
    sendport('b');
  }
}


//++++++++++++++++++++++++++++++ move motor 2 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void motor2(int direction){
  length2 += direction;

  if(direction < 0){
    sendport('c');
  }
  else{
    sendport('d');
  }
}

//++++++++++++++++++++++++++++++ move motor 3 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void motor3(int direction){
  length3 += direction;

  if(direction < 0){
    sendport('e');
  }
  else{
    sendport('f');
  }
}



//++++++++++++++++++++++++++++++++++++++ MoveToCoordinates +++++++++++++++++++++++++++++++++++++++++++++++++++++
int MoveToCoordinates(long X, long Y, long Z, int Servo3New, int Servo4New, int Servo5New){
  int i = 0, i2 = 0;
  double length1New, length2New, length3New;
  double M1MountX, M1MountY, M1MountZ;
  double M2MountX, M2MountY, M2MountZ;
  double M3MountX, M3MountY, M3MountZ;
  double doStepM1, doStepM2, doStepM3;
  double doStepS1, doStepS2, doStepS3, doStepS4, doStepS5;
  double doStepCounterM1, doStepCounterM2, doStepCounterM3;
  double doStepCounterS1, doStepCounterS2, doStepCounterS3, doStepCounterS4, doStepCounterS5;
  long offsetX, offsetY, offsetZ;
  long tempX, tempY, tempZ;
  long stepDiff = 0;
  double alpha = 90.0, beta = 90.0;
  double alpha3;
  double alpha4;
  long barMountZ;
  int Servo1New, Servo2New;

  if(Servo3New > SERVO_3_MAX){
    Servo3New = SERVO_3_MAX;
  }
  if(Servo3New < SERVO_3_MIN){
    Servo3New = SERVO_3_MIN;
  }

  if(Servo4New > SERVO_4_MAX){
    Servo4New = SERVO_4_MAX;
  }
  if(Servo4New < SERVO_4_MIN){
    Servo4New = SERVO_4_MIN;
  }

  if(Servo5New > SERVO_5_MAX){
    Servo5New = SERVO_5_MAX;
  }
  if(Servo5New < SERVO_5_MIN){
    Servo5New = SERVO_5_MIN;
  }

  barMountZ = INITIAL_BAR_LENGTH * StepsPermm;

  offsetX = GRIPPER_OFFSET * StepsPermm;
  offsetY = -GRIPPER_LENGTH * StepsPermm;
  offsetZ = LEVER_S5 * StepsPermm;

  //calculate X/Y/Z offset caused by rotation of servos 3 and 4
  alpha3 = Servo3New - 90;
  alpha4 = Servo4New - 90;

  tempZ = offsetZ * cos(alpha4 * PI / 180.0) - offsetY * sin(alpha4 * PI / 180.0);
  tempY = offsetY * cos(alpha4 * PI / 180.0) + offsetZ * sin(alpha4 * PI / 180.0);

  offsetY = tempY;

  tempX = offsetX * cos(alpha3 * PI / 180.0) - offsetY * sin(alpha3 * PI / 180.0);
  tempY = offsetY * cos(alpha3 * PI / 180.0) + offsetX * sin(alpha3 * PI / 180.0);




  tempX -= GRIPPER_OFFSET * StepsPermm;
  tempY += GRIPPER_LENGTH * StepsPermm;
  tempY = -tempY;
  tempZ -= LEVER_S5 * StepsPermm;

  offsetX = X + tempX;
  offsetY = Y + tempY;
  offsetZ = Z + tempZ;




  if(offsetY != 0){
    alpha = atan((double)(barMountZ - offsetZ) / (double)(offsetY)) * 180.0 / PI;
  }

  if(offsetX != 0){
    beta = atan((double)((barMountZ - offsetZ) / offsetX)) * 180.0 / PI;
  }

  Servo1New = alpha;
  if(Servo1New < 0){
    Servo1New = -Servo1New;
  }
  else{
    Servo1New = 90 + (90 - Servo1New);
  }
  if(Servo1New > SERVO_1_MAX){
    Servo1New = SERVO_1_MAX;
  }
  if(Servo1New < SERVO_1_MIN){
    Servo1New = SERVO_1_MIN;
  }


  Servo2New = beta;
  if(Servo2New < 0){
    Servo2New = -Servo2New;
  }
  else{
    Servo2New = 90 + (90 - Servo2New);
  }
  if(Servo2New > SERVO_2_MAX){
    Servo2New = SERVO_2_MAX;
  }
  if(Servo2New < SERVO_2_MIN){
    Servo2New = SERVO_2_MIN;
  }

  M1MountX = offsetX - RADIUS_CORD_MOUNT * StepsPermm * cos(30.0 * PI / 180.0);
  M1MountY = offsetY + RADIUS_CORD_MOUNT * StepsPermm * sin(30.0 * PI / 180.0);;
  M1MountZ = offsetZ;

  M2MountX = offsetX;
  M2MountY = offsetY - RADIUS_CORD_MOUNT * StepsPermm;
  M2MountZ = offsetZ;

  M3MountX = offsetX + RADIUS_CORD_MOUNT * StepsPermm * cos(30.0 * PI / 180.0);
  M3MountY = offsetY + RADIUS_CORD_MOUNT * StepsPermm * sin(30.0 * PI / 180.0);;
  M3MountZ = offsetZ;

  length1New = sqrt(pow(M1X - M1MountX, 2.0) + pow(M1Y - M1MountY, 2.0) + pow(M1Z - M1MountZ, 2.0));
  length2New = sqrt(pow(M2X - M2MountX, 2.0) + pow(M2Y - M2MountY, 2.0) + pow(M2Z - M2MountZ, 2.0));
  length3New = sqrt(pow(M3X - M3MountX, 2.0) + pow(M3Y - M3MountY, 2.0) + pow(M3Z - M3MountZ, 2.0));

  stepDiff = abs(length1New - length1);
  if(abs(length2New - length2) > stepDiff){
    stepDiff = abs(length2New - length2);
  }
  if(abs(length3New - length3) > stepDiff){
    stepDiff = abs(length3New - length3);
  }
  if(abs(Servo1New - Servo1) > stepDiff){
    stepDiff = abs(Servo1New - Servo1);
  }
  if(abs(Servo2New - Servo2) > stepDiff){
    stepDiff = abs(Servo2New - Servo2);
  }
  if(abs(Servo3New - Servo3) > stepDiff){
    stepDiff = abs(Servo3New - Servo3);
  }
  if(abs(Servo4New - Servo4) > stepDiff){
    stepDiff = abs(Servo4New - Servo4);
  }
  if(abs(Servo5New - Servo5) > stepDiff){
    stepDiff = abs(Servo5New - Servo5);
  }

  doStepM1 = (double)(abs(length1New - length1)) / (double)(stepDiff);
  doStepM2 = (double)(abs(length2New - length2)) / (double)(stepDiff);
  doStepM3 = (double)(abs(length3New - length3)) / (double)(stepDiff);

  doStepS1 = (double)(abs(Servo1New - Servo1)) / (double)(stepDiff);
  doStepS2 = (double)(abs(Servo2New - Servo2)) / (double)(stepDiff);
  doStepS3 = (double)(abs(Servo3New - Servo3)) / (double)(stepDiff);
  doStepS4 = (double)(abs(Servo4New - Servo4)) / (double)(stepDiff);
  doStepS5 = (double)(abs(Servo5New - Servo5)) / (double)(stepDiff);

  doStepCounterM1 = 0.0;
  doStepCounterM2 = 0.0;
  doStepCounterM3 = 0.0;
  doStepCounterS1 = 0.0;
  doStepCounterS2 = 0.0;
  doStepCounterS3 = 0.0;
  doStepCounterS4 = 0.0;
  doStepCounterS5 = 0.0;

  for(i=0; i<stepDiff + 2; i++){
    doStepCounterM1 += doStepM1;
    doStepCounterM2 += doStepM2;
    doStepCounterM3 += doStepM3;

    doStepCounterS1 += doStepS1;
    doStepCounterS2 += doStepS2;
    doStepCounterS3 += doStepS3;
    doStepCounterS4 += doStepS4;
    doStepCounterS5 += doStepS5;

    if(doStepCounterM1 >= 1.0){
      if(length1New > length1){
        motor1(1);
      }
      if(length1New < length1){
        motor1(-1);
      }
      doStepCounterM1 -= 1.0;
    }

    if(doStepCounterM2 >= 1.0){
      if(length2New > length2){
        motor2(1);
      }
      if(length2New < length2){
        motor2(-1);
      }
      doStepCounterM2 -= 1.0;
    }

    if(doStepCounterM3 >= 1.0){
      if(length3New > length3){
        motor3(1);
      }
      if(length3New < length3){
        motor3(-1);
      }
      doStepCounterM3 -= 1.0;
    }

    if(doStepCounterS1 >= 1.0){
      if(Servo1New > Servo1){
        moveServo1(1);
      }
      if(Servo1New < Servo1){
        moveServo1(-1);
      }
      doStepCounterS1 -= 1.0;
    }

    if(doStepCounterS2 >= 1.0){
      if(Servo2New > Servo2){
        moveServo2(1);
      }
      if(Servo2New < Servo2){
        moveServo2(-1);
      }
      doStepCounterS2 -= 1.0;
    }

    if(doStepCounterS3 >= 1.0){
      if(Servo3New > Servo3){
        moveServo3(1);
      }
      if(Servo3New < Servo3){
        moveServo3(-1);
      }
      doStepCounterS3 -= 1.0;
    }

    if(doStepCounterS4 >= 1.0){
      if(Servo4New > Servo4){
        moveServo4(1);
      }
      if(Servo4New < Servo4){
        moveServo4(-1);
      }
      doStepCounterS4 -= 1.0;
    }

    if(doStepCounterS5 >= 1.0){
      if(Servo5New > Servo5){
        moveServo5(1);
      }
      if(Servo5New < Servo5){
        moveServo5(-1);
      }
      doStepCounterS5 -= 1.0;
    }

  }//for(i=0; i<stepDiff; i++){


  currentX = X;
  currentY = Y;
  currentZ = Z;



  return 0;
}
//-------------------------------------- MoveToCoordinates -----------------------------------------------

//++++++++++++++++++++++++++++++++++++++ KeyControl ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void KeyControl(int action){
int i;
int i2;

  i = 0;
  SingleKey = 1;
  KeyCode[0] = 0;
  KeyCode[1] = 0;
  KeyCode[2] = 0;
  KeyCode[3] = 0;
  KeyCode[4] = 0;
  KeyHit = 0;
  while (kbhit()){
    KeyHit = getch();
    KeyCode[i] = KeyHit;
    i++;
    if(i == 5){
      i = 0;
    }
    if(i > 1){
      SingleKey = 0;
    }
  }
  if(SingleKey == 0){
    KeyHit = 0;
  }

  if(action == 1 || action == 2){
    //Move X-axis
    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 68 && KeyCode[3] == 0 && KeyCode[4] == 0){
      if(initBot == 0){
        MoveToCoordinates(currentX - MoveLength, currentY, currentZ, Servo3, Servo4, Servo5);
      }
      else{
        for(i2=0; i2<MoveLength; i2++){
          motor1(-1);
          length1++;
        }
      }
    }

    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 67 && KeyCode[3] == 0 && KeyCode[4] == 0){
      if(initBot == 0){
        MoveToCoordinates(currentX + MoveLength, currentY, currentZ, Servo3, Servo4, Servo5);
      }
      else{
        for(i2=0; i2<MoveLength; i2++){
          motor1(1);
          length1--;
        }
      }
    }

    //Move Y-axis
    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 65 && KeyCode[3] == 0 && KeyCode[4] == 0){
      if(initBot == 0){
        MoveToCoordinates(currentX, currentY + MoveLength, currentZ, Servo3, Servo4, Servo5);
      }
      else{
        for(i2=0; i2<MoveLength; i2++){
          motor2(-1);
          length2++;
        }
      }
    }

    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 66 && KeyCode[3] == 0 && KeyCode[4] == 0){
      if(initBot == 0){
        MoveToCoordinates(currentX, currentY - MoveLength, currentZ, Servo3, Servo4, Servo5);
      }
      else{
        for(i2=0; i2<MoveLength; i2++){
          motor2(1);
          length2--;
        }
      }
    }

    //Move Z-axis
    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 53 && KeyCode[3] == 126 && KeyCode[4] == 0){
      if(initBot == 0){
        MoveToCoordinates(currentX, currentY, currentZ + MoveLength, Servo3, Servo4, Servo5);
      }
      else{
        for(i2=0; i2<MoveLength; i2++){
          motor3(-1);
          length3++;
        }
      }
    }

    if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 54 && KeyCode[3] == 126 && KeyCode[4] == 0){
      if(initBot == 0){
        MoveToCoordinates(currentX, currentY, currentZ - MoveLength, Servo3, Servo4, Servo5);
      }
      else{
        for(i2=0; i2<MoveLength; i2++){
          motor3(1);
          length3--;
        }
      }
    }

    if(KeyHit == 'l'){
      MoveToCoordinates(currentX, currentY, currentZ, Servo3 + MoveLength, Servo4, Servo5);
      //moveServo3(1);
    }
    if(KeyHit == 'r'){
      MoveToCoordinates(currentX, currentY, currentZ, Servo3 - MoveLength, Servo4, Servo5);
      //moveServo3(-1);
    }

    if(KeyHit == 'u'){
      MoveToCoordinates(currentX, currentY, currentZ, Servo3, Servo4 - MoveLength, Servo5);
      //moveServo4(-1);
    }
    if(KeyHit == 'd'){
      MoveToCoordinates(currentX, currentY, currentZ, Servo3, Servo4 + MoveLength, Servo5);
      //moveServo4(1);
    }

    if(KeyHit == 'o'){
      MoveToCoordinates(currentX, currentY, currentZ, Servo3, Servo4, Servo5 - MoveLength);
      //moveServo5(-1);
    }
    if(KeyHit == 'c'){
      MoveToCoordinates(currentX, currentY, currentZ, Servo3, Servo4, Servo5 + MoveLength);
      //moveServo5(1);
    }

    if(KeyHit == '0'){
      MoveToCoordinates(0, 0, 0, 90, 90, 90);
    }

    if(KeyHit == 'm'){
      switch (MoveLength){
        case 1:
          MoveLength = 4;
          break;
        case 4:
          MoveLength = 12;
          break;
        case 12:
          MoveLength = 48;
          break;
        case 48:
          MoveLength = 96;
          break;
        case 96:
          MoveLength = 1;
      }

      //PrintMenue_01();
    }
    if(KeyHit != 0 || SingleKey == 0){
      PrintMenue_01();
    }
  }


}
//-------------------------------------- KeyControl --------------------------------------------------------


//######################################################################
//################## Main ##############################################
//######################################################################

int main(int argc, char **argv){

  int MenueLevel = 0;
  char FullFileName[1000] = "";
  char FileNameOld[1000] = "";
  struct winsize terminal;
  int i;
  int FileSelected = 0;
  int FileStartRow = 0;
  char *pEnd;
  FILE *PlotFile;
  char TextLine[30000];
  long coordinateCount = 0;
  char a;
  int stopPlot = 0;
  long PlotStartTime = 0;



  //Calculate XYZ coordinates of motors
  M1X = -BASELENGTH / 2.0;
  M1Y = BASELENGTH / 2.0 * tan(30.0 * PI / 180.0);

  M2X = 0.0;
  M2Y = -(BASELENGTH * cos(30.0 * PI / 180.0) - M1Y);

  M3X = BASELENGTH / 2.0;
  M3Y = BASELENGTH / 2.0 * tan(30.0 * PI / 180.0);

  M2Z = sqrt(pow(INITIAL_LENGTH, 2.0) - pow(M2Y + RADIUS_CORD_MOUNT, 2.0));
  M1Z = M2Z;
  M3Z = M2Z;

  M1X = M1X * StepsPermm;
  M1Y = M1Y * StepsPermm;
  M1Z = M1Z * StepsPermm;

  M2X = M2X * StepsPermm;
  M2Y = M2Y * StepsPermm;
  M2Z = M2Z * StepsPermm;

  M3X = M3X * StepsPermm;
  M3Y = M3Y * StepsPermm;
  M3Z = M3Z * StepsPermm;

  printf("M1X = %lf, M1Y = %lf, M1Z = %lf\n", M1X, M1Y, M1Z);
  printf("M2X = %lf, M2Y = %lf, M2Z = %lf\n", M2X, M2Y, M2Z);
  printf("M3X = %lf, M3Y = %lf, M3Z = %lf\n", M3X, M3Y, M3Z);

  length1 = INITIAL_LENGTH * StepsPermm;
  length2 = INITIAL_LENGTH * StepsPermm;
  length3 = INITIAL_LENGTH * StepsPermm;

  printf("\n\nKoordinates in steps:\n");
  printf("M1X = %lf, M1Y = %lf, M1Z = %lf\n", M1X, M1Y, M1Z);
  printf("M2X = %lf, M2Y = %lf, M2Z = %lf\n", M2X, M2Y, M2Z);
  printf("M3X = %lf, M3Y = %lf, M3Z = %lf\n", M3X, M3Y, M3Z);



  openport();

  printf("\n\nWaiting for 'X' from Arduino (Arduino pluged in?)...\n");

  //Wait for 'X' from Arduino
  while(readport() != 'X');



  strcpy(FileName, "noFiLE");

  getcwd(ReplayPath, 1000);
  strcat(ReplayPath, "/replay-files");
  printf("ReplayPath=>%s<", ReplayPath);

  if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &terminal)<0){
    printf("Can't get size of terminal window");
  }
  else{
    MaxRows = terminal.ws_row;
    MaxCols = terminal.ws_col;
    MessageY = MaxRows-3;
  }

  clrscr(1, MaxRows);
  PrintRow('-', MessageY - 1);
  PrintMenue_01();

  MaxFileRows = MaxRows - 10;

  while (1){
    MessageText("Waiting for key press.", MessageX, MessageY, 0);

    if(MenueLevel == 0){
      KeyControl(1);

      if(KeyHit == 'f'){
        strcpy(FileNameOld, FileName);
        strcpy(FileName, PrintMenue_02(FileStartRow, FileSelected));
        MenueLevel = 1;
      }

      if(KeyHit == 's'){
        sprintf(TextLine, "echo \"% 7ld, % 7ld, % 7ld, % 7d, % 7d, % 7d,        0;\" >> coordinates.bot", currentX, currentY, currentZ, Servo3, Servo4, Servo5);
        system(TextLine);
        sprintf(TextLine, "X=%ld, Y=%ld, Z=%ld, S3=%d, S4=%d, S5=%d stored in coordinates.bot.", currentX, currentY, currentZ, Servo3, Servo4, Servo5);
        MessageText(TextLine, MessageX, MessageY, 0);
      }

      if(KeyHit == 'i'){
        if(initBot == 0){
          initBot = 1;
        }
        else{
      currentX = 0;
      currentY = 0;
      currentZ = 0;
          initBot = 0;
          MoveToCoordinates(0, 0, 0, 90, 90, 90);
        }
        PrintMenue_01();
      }//if(KeyHit == 'i'){


      if(KeyHit == 'p'){//Plot file
        int coordReadNo = 0;
        long goToX, goToY, goToZ;
        int goToS3, goToS4, goToS5;
        int sleepTime;
        int escPress = 0;
        if(kbhit()){
          getch();
          KeyHit = 27;
        }
        if(strcmp(FileName, "noFiLE") != 0){
          if((PlotFile=fopen(FullFileName,"rb"))==NULL){
            sprintf(TextLine, "Can't open file '%s'!\n", FullFileName);
            strcpy(FileName, "NoFiLE");
            ErrorText(TextLine);
          }
        }
        if(strcmp(FileName, "noFiLE") != 0 && KeyHit != 27){
          PlotStartTime = time(0);
          coordinateCount = 0;
          stopPlot = 0;

          while(!(feof(PlotFile)) && stopPlot == 0){
            fread(&a, 1, 1, PlotFile);
            i=0;
            TextLine[0] = '\0';
            while(!(feof(PlotFile)) && a !=' ' && a != ';' && a != '\"' && a != ',' && a != 10 && a!= 13){
              TextLine[i] = a;
              TextLine[i+1] = '\0';
              i++;
              fread(&a, 1, 1, PlotFile);
            }
            if(a == '#'){
              while(!(feof(PlotFile)) && a != 10 && a!= 13){
                fread(&a, 1, 1, PlotFile);
              }
            }
            if(a == ',' || a == ';'){//Coordinates found
              coordReadNo++;
              if(coordReadNo == 1){
                goToX = strtol(TextLine, &pEnd, 10);
              }
              if(coordReadNo == 2){
                goToY = strtol(TextLine, &pEnd, 10);
              }
              if(coordReadNo == 3){
                goToZ = strtol(TextLine, &pEnd, 10);
              }
              if(coordReadNo == 4){
                goToS3 = strtol(TextLine, &pEnd, 10);
              }
              if(coordReadNo == 5){
                goToS4 = strtol(TextLine, &pEnd, 10);
              }
              if(coordReadNo == 6){
                goToS5 = strtol(TextLine, &pEnd, 10);
              }
              if(coordReadNo == 7){
                sleepTime = strtol(TextLine, &pEnd, 10);
                sprintf(TextLine, "X=%ld, Y=%ld, Z=%ld, S3=%d, S4=%d, S5=%d, sleep=%ds", goToX, goToY, goToZ, goToS3, goToS4, goToS5, sleepTime);
                MessageText(TextLine, MessageX, MessageY, 0);
                coordReadNo = 0;
                MoveToCoordinates(goToX, goToY, goToZ, goToS3, goToS4, goToS5);
                sleep(sleepTime);
                //sleep(3);// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                coordinateCount++;
                PrintMenue_03(FullFileName, coordinateCount, PlotStartTime);
                while(kbhit()){
                  escPress = getch();
                }
                if(escPress == 27){
                  while(!(feof(PlotFile))){
                    fread(&a, 1, 1, PlotFile);
                  }
                }
              }

            }
          }//while(!(feof(PlotFile)) && stopPlot == 0){
          fclose(PlotFile);
          PrintMenue_03(FullFileName, coordinateCount, PlotStartTime);
          while(kbhit()){
            getch();
          }
          MessageText("Finished! Press any key to return to main menu.", MessageX, MessageY, 0);
          getch();
          PrintMenue_01();

        }//if(strcmp(FileName, "noFiLE") != 0){
      }//if(KeyHit == 'p'){


    }//if(MenueLevel == 0){

    if(MenueLevel == 1){//Select file
      KeyControl(0);
      if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 66 && KeyCode[3] == 0 && KeyCode[4] == 0){
        if(FileSelected < FilesFound - 1){
          FileSelected++;
          if(FileSelected > MaxFileRows - 2){
            FileStartRow = FileSelected - MaxFileRows + 2;
          }
          strcpy(FileName, PrintMenue_02(FileStartRow, FileSelected));
        }
      }

      if(KeyCode[0] == 27 && KeyCode[1] == 91 && KeyCode[2] == 65 && KeyCode[3] == 0 && KeyCode[4] == 0){
        if(FileSelected > 0){
          if(FileSelected == FileStartRow + 1){
            if(FileStartRow > 0){
              FileStartRow--;
            }
          }
          FileSelected--;
          strcpy(FileName, PrintMenue_02(FileStartRow, FileSelected));
        }
      }

      if(KeyHit == 10){//Set selected file
        MenueLevel = 0;
        clrscr(MessageY + 1, MessageY + 1);
        strcpy(FullFileName, ReplayPath);
        strcat(FullFileName, "/");
        strcat(FullFileName, FileName);
        if((PlotFile=fopen(FullFileName,"rb"))==NULL){
          sprintf(TextLine, "Can't open file '%s'!\n", FullFileName);
          ErrorText(TextLine);
          strcpy(FileName, "NoFiLE");
        }
        PrintMenue_01();
      }//if(KeyHit == 10){

    }//if(MenueLevel == 1){


    if(KeyHit == 27){
      if(MenueLevel == 0){
        clrscr(MessageY + 1, MessageY + 1);
        PrintYN();
        //MessageText("Exit program (y/n)?", MessageX, MessageY + 1, 0);
        while(KeyHit != 'y' && KeyHit != 'n'){
          KeyHit = getch();
          if(KeyHit == 'y'){
            exit(0);
          }
        }
        PrintMenue_01();
      }
      if(MenueLevel == 1){
        MenueLevel = 0;
        strcpy(FileName, FileNameOld);
        PrintMenue_01();
      }
      clrscr(MessageY + 1, MessageY + 1);
    }
  }

  return 0;
}
